/************************************************************************\
**  ftp.c - File Transfert Protocol                                     **
**                                                                      **
**  Copyright (c) 2004 Christophe Blaess <ccb@club-internet.fr>         **
**    ---------------------------------------------------------------   **
**                                                                      **
** This program is free software; you can redistribute it and/or modify **
** it under the terms of the GNU General Public License as published by **
** the Free Software Foundation.                                        **
**                                                                      **
**  This program is distributed in the hope that it will be useful,     **
** but WITHOUT ANY WARRANTY; without even the implied warranty of       **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **
** GNU General Public License for more details.                         **
**                                                                      **
** You should have received a copy of the GNU General Public License    **
** along with this program; if not, write to the Free Software          **
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA             **
** 02111-1307  USA                                                      **
**                                                                      **
**    ---------------------------------------------------------------   **
**                                                                      **
** Ce programme est libre, vous pouvez le redistribuer et/ou le modifier**
** selon les termes de la Licence Publique Gnrale GNU publie par la  **
** Free Software Foundation.                                            **
**                                                                      **
** Ce programme est distribu car potentiellement utile, mais SANS      **
** AUCUNE GARANTIE, ni explicite ni implicite, y compris les garanties  **
** de commercialisation ou d'adaptation dans un but spcifique.         **
** Reportez-vous  la Licence Publique Gnrale GNU pour plus de dtails**
**                                                                      **
** Vous devez avoir reu une copie de la Licence Publique Gnrale GNU  **
** en mme temps que ce programme ; si ce n'est pas le cas, crivez    **
** la Free Software Foundation, Inc, 59 Temple Place, Suite 330, Boston **
** MA 02111-1307, tats-Unis.                                           **
**                                                                      **
\************************************************************************/

	/* See RFC 959 for information on FTP
	   extension RFC 775
	 */

	#include <ctype.h>
	#include <dirent.h>
	#include <errno.h>
	#include <getopt.h>
	#include <signal.h>
	#include <stdarg.h>
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>
	#include <unistd.h>
	#include <arpa/inet.h>
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <sys/fcntl.h>

	#include "multiserv.h"
	#include "multiserv_utils.h"

	#include "ftp.h"

#if defined(MULTISERV_FTP_TCP)

	static int		ftp_dialog (int sock, struct sockaddr_in * unused_client_address);


	static char *	ftp_getcwd(void);

	static int		ftp_check_access_setuid_chdir	(char * login, char * password);
	static int		get_destination_address			(struct sockaddr_in *address, char * string);
	static int		ftp_send_list					(int control_socket, struct sockaddr_in * destination, char * directory);


	int
ftp_description (FILE * stream)
{
	return (fprintf(stream, "File Transfert Protocol (FTP) rfc.959 (TCP/IP)"));
}


	int
ftp_main (int argc, char * argv[])
{
	multiserv_log(MULTISERV_LOG_MESSAGE, "ftp", "TCP server start");
	return (tcp_server(ftp_dialog));
}


	static int
ftp_send(int sock, char * string)
{
	int n = strlen(string);
	if (send(sock, string, n, 0) == n)
		return (0);
	return (-1);
}

	static int
ftp_wait_string (int sock)
{
	int n;
	do {
		n = recv(sock, multiserv_buffer,MULTISERV_BUFFER_LEN, 0);
		if (n <= 0)
			return (-1);
		do {
			multiserv_buffer[n] = '\0';
			n --;
		} while ((n>=0) && (isspace(multiserv_buffer[n])));
	} while (n < 0);
	return (0);
}

	static int
ftp_dialog (int sock, struct sockaddr_in * client_address)
{
	char *	arg = NULL;
	char *	logname = NULL;
	char *	pathname = NULL;
	int		access_granted = 0;
	int		fd;
	struct	sockaddr_in	destination;
	
	multiserv_debug("ftp", "Got a connection, sending banner");	
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send(sock, "220 Multiserv FTP server ready\r\n");
	#else
		ftp_send(sock, "220 \r\n");
	#endif	
	memset(&destination, 0, sizeof(struct sockaddr_in));
	
	while(1) {
	
		/* ## TODO : add a timeout */
		if (ftp_wait_string(sock) != 0)
			break;

		arg = multiserv_buffer;
		while ((arg[0] != '\0') && (arg[0] != ' '))
			arg ++;
		while(arg[0] == ' ')
			arg ++;

		/* QUIT is always available */
		if (strncasecmp (multiserv_buffer, "QUIT", 4) == 0) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send(sock, "221 Bye.\r\n");
			#else
				ftp_send(sock, "221 \r\n");
			#endif
			multiserv_debug("ftp", "Connection closed by client");
			break;
		}
		
		/* USER / PASS is only allowed when access is NOT already granted,
		   because we chroot during some connections */
		if (! access_granted) {
			if (strncasecmp (multiserv_buffer, "USER ", 5) == 0) {
				if (logname != NULL)
					free (logname);
				logname = malloc (strlen(arg) + 1);
				if (logname == NULL) {
					#ifdef MULTISERV_FTP_VERBOSE
						ftp_send(sock, "421 Service not available\r\n");
					#else
						ftp_send(sock, "421 \r\n");
					#endif
					break;
				}
				strcpy(logname, arg);
				#ifdef MULTISERV_FTP_ANONYMOUS
					if (strcmp(logname, "anonymous") == 0)
						#ifdef MULTISERV_FTP_VERBOSE
							ftp_send(sock, "331 Please send email adress as password\r\n");
						#else
							ftp_send(sock, "331 \r\n");
						#endif
					else
				#endif
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "331 User name ok, need password.\r\n");
				#else
					ftp_send(sock, "331 \r\n");
				#endif
				continue;
			}

			if (strncasecmp(multiserv_buffer, "PASS ", 5) == 0) {
				access_granted = 0;
				if (logname == NULL) {
					#ifdef MULTISERV_FTP_VERBOSE
						ftp_send(sock, "503 Please log in with USER first\r\n");
					#else
						ftp_send(sock, "503 \r\n");
					#endif
					continue;
				}
				#ifdef MULTISERV_FTP_ANONYMOUS
					if (strcmp(logname, "anonymous") == 0) {
						/* Anonymous ftp is only allowed for server ran by root */
						if ((chdir(MULTISERV_FTP_ANONYMOUS_DIR) != 0)
						 || (chroot(MULTISERV_FTP_ANONYMOUS_DIR) != 0)
						 || (setgid(MULTISERV_FTP_ANONYMOUS_GID) != 0)
						 || (setuid(MULTISERV_FTP_ANONYMOUS_UID) != 0)) {
							#ifdef MULTISERV_FTP_VERBOSE
								ftp_send(sock, "421 Service not available\r\n");
							#else
								ftp_send(sock, "421 \r\n");
							#endif
							break;
						}
						#ifdef MULTISERV_FTP_VERBOSE
							ftp_send(sock, "230 Welcome A.N.Onymous.\r\n");
						#else
							ftp_send(sock, "230 \r\n");
						#endif
						/* ## log connection */
						access_granted = 1;
						continue;
					}
				#endif
				access_granted = ftp_check_access_setuid_chdir(logname, arg);
				if (access_granted)
					#ifdef MULTISERV_FTP_VERBOSE
						ftp_send(sock, "230 User logged in, proceed\r\n");
					#else
						ftp_send(sock, "230 \r\n");
					#endif
				else
					#ifdef MULTISERV_FTP_VERBOSE
						ftp_send(sock, "530 Login incorrect\r\n");
					#else
						ftp_send(sock, "530 \r\n");
					#endif
				memset(multiserv_buffer, 0, MULTISERV_BUFFER_LEN); /* paranoia */
				continue;
			}
			
			/* All other commands need preliminary login */
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send(sock, "530\r\n");
			#else
				ftp_send(sock, "530 You need to log in first!\r\n");
			#endif
			continue;
		}
		
		if ((strcasecmp(multiserv_buffer, "NOOP") == 0) 
		||  (strncasecmp(multiserv_buffer, "ALLO", 4) == 0)) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send(sock, "200 Ok\n\n");
			#else
				ftp_send(sock, "200 \n\n");
			#endif
			continue;
		}
	
		if (strcasecmp(multiserv_buffer, "SYST") == 0) {
			/* ## TODO: is there any compile-time constant giving the OS name ? */
				ftp_send(sock, "215 LINUX\r\n");
			continue;
		}

		if ((strcasecmp(multiserv_buffer, "PWD") == 0)
		 || (strcasecmp(multiserv_buffer, "XPWD") == 0)) {
		 	pathname = ftp_getcwd();
			if (pathname == NULL) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "451 Local error\r\n");
				#else
					ftp_send(sock, "451 \r\n");
				#endif
				break;
			}
			if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "257 \"%s\"\r\n", pathname) >= MULTISERV_BUFFER_LEN) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "451 Local error\r\n");
				#else
					ftp_send(sock, "451 \r\n");
				#endif
				free(pathname);
				pathname = NULL;
				break;
			}
			ftp_send(sock, multiserv_buffer);
			free(pathname);
			pathname = NULL;
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "CWD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XCWD ", 5) == 0)) {
			if (chdir(arg) != 0)
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "550 Unable to change directory\r\n");
				#else
					ftp_send(sock, "550 \r\n");
				#endif
			else
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "250 Directory Ok\r\n");
				#else
					ftp_send(sock, "250 \r\n");
				#endif
			continue;
		}

		if ((strcasecmp(multiserv_buffer, "CDUP") == 0)
		|| (strcasecmp(multiserv_buffer, "XCUP") == 0)) {
			if (chdir("..") != 0)
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "550 Unable to change directory\r\n");
				#else
					ftp_send(sock, "550 \r\n");
				#endif
			else
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "250 Directory Ok\r\n");
				#else
					ftp_send(sock, "250 \r\n");
				#endif
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "MKD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XMKD ", 5) == 0)) {
			if (mkdir(arg, 0755) != 0) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "550 Creation failed\r\n");
				#else
					ftp_send(sock, "550 d\r\n");
				#endif
				continue;
			}
			if ((fd = open (".", O_RDONLY)) < 0) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "451 Local error\r\n");
				#else
					ftp_send(sock, "451 \r\n");
				#endif
				break;
			}
			if ((chdir(arg) != 0)
			 || ((pathname = ftp_getcwd()) == NULL)) {
			 	fchdir(fd);
			 	close(fd);
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "451 Local error\r\n");
				#else
					ftp_send(sock, "451 \r\n");
				#endif
				break;
			}
			if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "257 \"%s\"\r\n", pathname) >= MULTISERV_BUFFER_LEN) { 
				fchdir(fd);
				close(fd);
				free(pathname);
				pathname = NULL;
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "451 Local error\r\n");
				#else
					ftp_send(sock, "451 \r\n");
				#endif
				break;
			}
			fchdir(fd);
			close(fd);
			free(pathname);
			pathname = NULL;
			ftp_send(sock, multiserv_buffer);
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "RMD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XRMD ", 5) == 0)) {
			if (rmdir(arg) != 0)
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "550 Delete failed\r\n");
				#else
					ftp_send(sock, "550 \r\n");
				#endif
			else
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "250 Delete ok\r\n");
				#else
					ftp_send(sock, "250 \r\n");
				#endif
			continue;
		}

		if (strncasecmp(multiserv_buffer, "DELE ", 5) == 0) {
			if (unlink(arg) != 0)
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "550 Delete failed\r\n");
				#else
					ftp_send(sock, "550 \r\n");
				#endif
			else
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "250 Delete ok\r\n");
				#else
					ftp_send(sock, "250 \r\n");
				#endif
			continue;
		}

		if (strncasecmp(multiserv_buffer, "PORT ", 5) == 0) {
			if (get_destination_address(& destination, multiserv_buffer+5) < 0) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send(sock, "500 Command unknown\r\n");
				#else
					ftp_send(sock, "500 \r\n");
				#endif
				continue;
			}
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send(sock, "200 PORT Command ok\r\n");
			#else
				ftp_send(sock, "200 \r\n");
			#endif
			continue;
		}

		if ((strncasecmp(multiserv_buffer,"LIST", 4) == 0)
		 || (strncasecmp(multiserv_buffer,"NLST", 4) == 0)) {
			if (ftp_send_list(sock, & destination, arg) != 0)
				break;
			continue;
		}
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(sock, "500 Command unknown\r\n");
		#else
			ftp_send(sock, "500 \r\n");
		#endif
fprintf(stderr,"UNKNOWN : %s\n", multiserv_buffer);		
	}
	if (logname != NULL)
		free(logname);
	logname = NULL;
	return (0);
}

	static int
ftp_check_access_setuid_chdir(char * login, char * password)
{
	/* hmmm, so, what we gonna do ? */
	
	/* lazy... */
	return (1);
}

	static char *
ftp_getcwd(void)
{
	char * pathname = NULL;
	char * new;
	int	size = 0;
	
	do {
		size += 4096;
		new = realloc (pathname, size);
		if (new == NULL) {
			if (pathname != NULL)
				free(pathname);
			return (NULL);
		} 
		pathname = new;
	} while (getcwd(pathname, size) == NULL);
	return (pathname);
}

	static int
get_destination_address(struct sockaddr_in *address, char * string)
{
	int a1, a2, a3, a4;
	int p1, p2;
	if (sscanf(string, "%d,%d,%d,%d,%d,%d", & a1, & a2, & a3, & a4, & p1, & p2) != 6)
		return (-1);
	if ((a1<0) || (a2<0) || (a3<0) || (a4<0) || (a1>255) || (a2>255) || (a3>255) || (a4>255))
		return (-1);
	memset(address, 0, sizeof(struct sockaddr_in));
	address->sin_family = AF_INET;
	address->sin_port = htons(p1*256+p2);
	address->sin_addr.s_addr = htonl(a1*16777216+a2*65536+a3*256+a4);
	return(0);
}


	static int
ftp_send_list(int control_socket, struct sockaddr_in * destination, char * directory)
{
	DIR * dp;
	struct dirent * entry;
	struct sockaddr_in source;
	
	int	data_socket;
	int	sockopt;
	memset(& source, 0, sizeof(struct sockaddr_in));
	source.sin_family = AF_INET;
	source.sin_addr.s_addr = htonl(INADDR_ANY);
	source.sin_port = htons (20);
	
	
	data_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (data_socket < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(control_socket, "451 Local error\r\n");
		#else
			ftp_send(control_socket, "451 \r\n");
		#endif
		return (1);
	}
	sockopt = 1;
	setsockopt(data_socket, SOL_SOCKET, SO_REUSEADDR, (char *) & sockopt, sizeof(sockopt));

	if (bind(data_socket, (struct sockaddr *) & source, sizeof(struct sockaddr_in)) < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(control_socket, "451 Local Error.\r\n");
		#else
			ftp_send(control_socket, "451 \r\n");
		#endif 
		close(data_socket);
		return (1);
	}
	
	alarm(2);
	if (connect(data_socket, (struct sockaddr *) destination, sizeof(struct sockaddr_in)) < 0) {
		alarm(0);
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(control_socket, "425 Data connection refused.\r\n");
		#else
			ftp_send(control_socket, "425 \r\n");
		#endif 
		close(data_socket);
		return (0);
	}
	alarm(0);
	if (directory[0] == '\0')
		dp = opendir(".");
	else
		dp = opendir(directory);
	if (dp == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(control_socket, "550 No such file or directory\r\n");
		#else
			ftp_send(control_socket, "550 \r\n");
		#endif
		close(data_socket);
		return (0);
	}
	while ((entry = readdir(dp)) != NULL) {
		if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "%s\r\n", entry->d_name) >= MULTISERV_BUFFER_LEN)
			continue;
		if (write(data_socket, multiserv_buffer, strlen(multiserv_buffer)) != strlen(multiserv_buffer))
				break;
	}
	close(data_socket);
	closedir(dp);
	if (entry != NULL) { /* error in ftp_send */
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send(control_socket, "451 Local error\r\n");
		#else
			ftp_send(control_socket, "451 \r\n");
		#endif
		return (1);
	}
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send(control_socket, "226 Data transfert complete\r\n");
	#else
		ftp_send(control_socket, "226 \r\n");
	#endif
	return (0);
}


#endif
